# -*- coding: utf-8 -*-

from django.core.mail.backends.smtp import EmailBackend

from ..models import ThxEmailBackend


class ThxEmailBackendMixin:

    def __init__(self, *args, **kwargs):
        self.environment_settings = self.get_environment_settings()
        return super().__init__(*args, **kwargs)

    @staticmethod
    def get_environment_settings():
        return ThxEmailBackend.get_instance_form_environment()

    def _log(self, message: str):
        return self.environment_settings._log(message)

    @staticmethod
    def search_differences(list_to_filter: list, list_to_compare: list) -> list:
        return list(filter(
            lambda address: address not in list_to_compare, list_to_filter
        ))

    def is_allowed_address(self, address):
        return address in self.environment_settings.allowed_addresses

    def is_forbidden_address(self, address):
        return address in self.environment_settings.forbidden_addresses

    def is_allowed_domain(self, domain):
        return domain in self.environment_settings.allowed_domains

    def is_forbidden_domain(self, domain):
        return domain in self.environment_settings.forbidden_domains

    def manage_logs(self, email_message, excluded_address: list):
        if excluded_address:
            self._log(
                f"{self.environment_settings.name}({self.environment_settings.environment}) - "
                f"the following address were excluded by settings, to: {excluded_address} - "
                f"email from: {email_message.from_email} \n"
            )
        else:
            self._log(
                f"{self.environment_settings.name}({self.environment_settings.environment}) - "
                f"sent an email to: {email_message.to} - "
                f"email from: {email_message.from_email} \n"
            )


class ThxEmailBackendEngine(ThxEmailBackendMixin, EmailBackend):
    """
    this Backend is thought to take track of dev emails sends around

    this class require the following settings:

    allowed_addresses: list of strings of allowed emails addresses
    forbidden_addresses: list of strings of forbidden emails addresses
    default_to_received_addresses: a default emails addresses allowed to receive all emails
    default_cc_received_addresses: a default emails addresses allowed to receive all CC emails
    default_bcc_received_addresses: a default emails addresses allowed to receive all BCC emails

    """

    def _send(self, email_message):
        email_message, excluded_address = self._clean_email_message(email_message)
        self.manage_logs(email_message, excluded_address)
        super()._send(email_message)

    def _check_allowed_address(self, addresses: list) -> tuple:
        """
        :param addresses: list of emails addresses string ex. ["code@thux.it" ...]
        :return: tuple of list of string of addresses allowed to receive platform emails.
        """
        allowed_address = []
        excluded_address = []
        for address in addresses:
            domain = address.split("@")[1]

            if self.is_allowed_address(address):
                allowed_address.append(address)
            elif self.is_forbidden_address(address):
                excluded_address.append(address)
            elif self.is_forbidden_domain(domain):
                excluded_address.append(address)
            elif self.is_allowed_domain(domain):
                allowed_address.append(address)
            else:
                if self.environment_settings.restraint_logic:
                    excluded_address.append(address)
                else:
                    allowed_address.append(address)
        return allowed_address, excluded_address

    def _clean_email_message(self, email_message) -> tuple:
        """
        :param email_message: EmailMessage object
        :return: EmailMessage object
        """
        excluded_address = []
        if email_message.to:
            email_message_to,  excluded_address_to = self._check_allowed_address(email_message.to)
            differences = self.search_differences(
                self.environment_settings.default_to_received_addresses, email_message_to
            )
            # if self.environment_settings.default_to_received_addresses not in email_message_to:
            if differences:
                # email_message_to.extend(self.environment_settings.default_to_received_addresses)
                email_message_to.extend(differences)
            email_message.to = email_message_to
            excluded_address.extend(excluded_address_to)

        if email_message.cc:
            email_message_cc, excluded_address_cc = self._check_allowed_address(email_message.cc)
            differences = self.search_differences(
                self.environment_settings.default_cc_received_addresses, email_message_cc
            )
            if differences:
                email_message_cc.extend(differences)
            email_message.cc = email_message_cc
            excluded_address.extend(excluded_address_cc)

        if email_message.bcc:
            email_message_bcc, excluded_address_bcc = self._check_allowed_address(email_message.bcc)
            differences = self.search_differences(
                self.environment_settings.default_bcc_received_addresses, email_message_bcc
            )
            if differences:
                email_message_bcc.extend(differences)
            email_message.bcc = email_message_bcc
            excluded_address.extend(excluded_address_bcc)

        return email_message, excluded_address
