# -*- coding: utf-8 -*-

import logging

from django.conf import settings
from django.contrib.postgres.fields import ArrayField
from django.core import checks
from django.core.exceptions import MultipleObjectsReturned, ValidationError
from django.db import models
from django.utils.translation import gettext_lazy as _
from thx.core.models import CleanModel, DateModel, OrderedModel, StatusModel, UserModel

logger = logging.getLogger('mail')


class ThxEmailBackend(CleanModel, UserModel, DateModel, StatusModel, OrderedModel):
    """
    Lo scopo di questo backed è LIMITARE i destinatari e l'uscita delle email dalla piattaforma che lo abilita.
    Nuova logica:
    Per ogni email da inviare vengono controllati gli indirizzi di destinazione contenuti ne campi to cc e bcc

    per ogni address contenuto nell'email, questo verrà validato come indirizzo uscente in base alla seguente logica
    a cascata decrescente di priorità:

    se l'indirizzo fa parte del campo allowed_address verrà sempre concessa l'inclusione tra i destinatari
    se l'indirizzo fa parte del campo forbidden_address non verrà concessa l'inclusione in nessun caso tra i destinatari
    se l'indirizzo ha un dominio che fa parte del campo forbidden_domain non verrà incluso in nessun caso tra i
    destinatari *
    se l'indirizzo ha un dominio che fa parte del campo allowed_domain verrà incluso tra i destinatari **
    se l'indirizzo o il suo dominio non sono elencati in nessuno dei campi di questo modello, essi non verranno inclusi
    tra i destinatari  a meno che non sia disabilitata la restraint_logic

    restraint_logic disabilitata:
    se l'indirizzo o il suo dominio non sono elencati in nessuno dei campi del modello, questo verrà sempre incluso tra
    i destinatari.

    *( tranne se l'indirizzo specifico fa parte dell'allowed_address)
    **( tranne se l'indirizzo specifico fa parte dell'forbidden_address)

    al momento i domini devo essere specificati per ogni top level domain, Ex: thux.it, thux.com, thux.org

    La logica dei log consiste nella possibilità di settare e abilitare/disabilitare i dei log aggiuntivi oltre
    ai cosueti log di errore

    """

    name = models.CharField(
        max_length=255,
        verbose_name=_('name')
    )
    environment = models.CharField(
        max_length=255,
        # choices=settings.THXEMAILBACKEND_THXEMAILBACKEND_ENVIRONMENT_CHOICES, # old logic
        verbose_name=_('environment'),
        help_text=_("environment ex: prod, staging, dev etc"),
    )
    allowed_addresses = ArrayField(
        models.EmailField(),
        default=list,
        null=True, blank=True,
        verbose_name=_('allowed addresses'),
        help_text=_("addresses allowed to receive emails from this backend"),  # always allowed?!
    )
    forbidden_addresses = ArrayField(
        models.EmailField(),
        default=list,
        null=True, blank=True,
        verbose_name=_('forbidden addresses'),
        help_text=_("addresses not allowed to receive emails from this backend"),
    )
    allowed_domains = ArrayField(
        # eventualmente per non specificare tutti i top level domain si potrebbe implementare una
        # logica a tuple ("thux", "__all"__,) per tutti e ("thux", "it",) per la specifica
        models.CharField(max_length=255),
        # https://www.rfc-editor.org/errata/eid1690#:~:text=It%20should%20say%3A-,In%20addition%20to%20restrictions%20on%20syntax%2C%20there%20is%20a%20length,total%20length%20of%20320%20characters.
        default=list,
        null=True, blank=True,
        verbose_name=_('allowed domains'),
        help_text=_("domains allowed to receive emails from this backend"),
    )
    forbidden_domains = ArrayField(
        models.CharField(max_length=255),
        default=list,
        null=True, blank=True,
        verbose_name=_('forbidden domains'),
        help_text=_("domains not allowed to receive emails from this backend"),
    )
    default_to_received_addresses = ArrayField(
        models.CharField(max_length=255),
        default=list,
        null=True, blank=True,
        verbose_name=_("default 'TO' recipient addresses"),
        help_text=_(
            "fill this field if you want these addresses to always be inserted in the 'TO' field of every email"
        ),
    )
    default_cc_received_addresses = ArrayField(
        models.CharField(max_length=255),
        default=list,
        null=True, blank=True,
        verbose_name=_("default 'CC' recipient addresses"),
        help_text=_(
            "fill this field if you want these addresses to always be inserted in the 'CC' field of every email"
        ),
    )
    default_bcc_received_addresses = ArrayField(
        models.CharField(max_length=255),
        default=list,
        null=True, blank=True,
        verbose_name=_('default bcc received addresses'),
        help_text=_(
            "fill this field if you want these addresses to always be inserted in the 'BCC' field of every email"
        ),
    )
    restraint_logic = models.BooleanField(
        default=True,
        verbose_name=_('restraint logic'),
        help_text=_(
            "if enabled: addresses or domains that are not included in one of the allowed or default fields "
            "will not be included in the recipients. <br>"
            "if disabled: every address or domain that is not included in one of the forbidden fields "
            "will always be allowed in the recipients."
        ),
    )
    """------------down there could be a new model in fk logic---------------------"""
    """we do not use choices to avoid migrations after added new logger"""
    enable_loggers = models.BooleanField(
        default=False,
        verbose_name=_('enable logger'),
        help_text=_("if enabled, logger writes information for every single email"),
    )
    logger = models.CharField(
        max_length=255,
        default='file',
        null=True, blank=True,
        verbose_name=_('logger'),
    )
    logger_level = models.CharField(
        max_length=255,
        default='INFO',
        null=True, blank=True,
        choices=settings.THXEMAILBACKEND_THXEMAILBACKEND_LOGGER_LEVELS_CHOICES,
        verbose_name=_('logger level'),
    )
    """------------------------------------------------------------------------------------"""

    class Meta:
        verbose_name = _('thx email backend')
        verbose_name_plural = _('thx email backends')
        unique_together = (('name', 'environment',),)
        ordering = ('ordering', 'environment', 'name', 'status')
        permissions = (
            ("list_thxemailbackend", "Can list thx email backend"),
            ("detail_thxemailbackend", "Can detail thx email backend"),
            ("enable_thxemailbackend", "Can enable thx email backend"),
            ("disable_thxemailbackend", "Can disable thx email backend"),
        )

    def __str__(self):
        return f'ThxEmailBackend: {self.name}'

    def save(self, *args, **kwargs):
        self.environment = self.environment.lower()
        self.set_status()
        self.set_allowed_addresses()
        self.set_forbidden_addresses()
        self.set_allowed_domains()
        self.set_forbidden_domains()
        self.set_default_to_received_addresses()
        self.set_default_cc_received_addresses()
        self.set_default_bcc_received_addresses()
        self.set_loggers()
        self.set_logger_level()
        super().save(*args, **kwargs)

    def clean(self):
        """write here your model validations and rise properly ValidationError({"field_name": "message"})"""
        # self.check_permissions()
        self.check_environment()
        self.check_addresses()
        self.check_domains()
        self.check_forbidden_addresses()
        self.check_logger()
        super().clean()

    def check_addresses(self):
        if self.allowed_addresses and self.forbidden_addresses:
            for address in self.allowed_addresses:
                if address in self.forbidden_addresses:
                    raise ValidationError({"allowed_addresses": _(
                        f"choose to set one among allowed or forbidden addresses for address: '{address}' "
                        f"you specified both"
                    )})

    def check_domains(self):
        if self.allowed_domains and self.forbidden_domains:
            for domain in self.allowed_domains:
                if domain in self.forbidden_domains:
                    raise ValidationError({"allowed_domains": _(
                        f"choose to set one among allowed or forbidden domain for domain: '{domain}' "
                        f"you specified both"
                    )})

    def check_environment(self):
        import glob
        envs = []
        for f_setting in glob.glob("web/settings/*.py"):
            with open(f_setting) as f:
                for line in f.readlines():
                    if 'EMAIL_BACKEND_ENVIRONMENT' in line:
                        try:
                            envs.append(line.split("=")[1].strip("\n  '").strip('"'))
                        except IndexError:
                            continue
        if self.environment not in envs:
            raise ValidationError({
                "environment": f"Environment: {self.environment} not found. "
                               f"You can create ThxEmailBackend instance only for environment that "
                               f"are already set into relative settings file. "
                               f"Check that your file {self.environment}.py exists and set EMAIL_BACKEND_ENVIRONMENT "
                               f"with the same name Ex: EMAIL_BACKEND_ENVIRONMENT = '{self.environment}'"
            })

    @staticmethod
    def search_duplicates(list_to_filter: list, list_to_compare: list, field_name: str):
        msg = "choose to set one among {field_name} or forbidden_addresses" \
              " for addresses:'{recipients}'. You specified both."
        duplicates = list(
            filter(lambda address: address in list_to_compare, list_to_filter)
        )
        if duplicates:
            raise ValidationError(
                {"forbidden_addresses": _(msg.format(field_name=field_name, recipients=duplicates))}
            )

    def check_forbidden_addresses(self):
        if self.forbidden_addresses:
            if self.default_to_received_addresses:
                self.search_duplicates(
                    self.forbidden_addresses, self.default_to_received_addresses, "default_to_received_addresses"
                )
            if self.default_cc_received_addresses:
                self.search_duplicates(
                    self.forbidden_addresses, self.default_cc_received_addresses, "default_cc_received_addresses"
                )
            if self.default_bcc_received_addresses:
                self.search_duplicates(
                    self.forbidden_addresses, self.default_bcc_received_addresses, "default_bcc_received_addresses"
                )

    def check_logger(self):
        if self.enable_loggers and self.logger:
            logger = self.logger.lower()
            from django.conf import settings
            loggers = settings.LOGGING.get("loggers", {})
            handlers = settings.LOGGING.get("handlers", {})
            if logger not in loggers.keys():
                raise ValidationError({
                    "logger": _(f"logger: {logger} does not exist look to your project settings to add it.")
                })
            admin_email_handler = None
            for handler, value in handlers.items():
                if value.get("class", "") == "django.utils.log.AdminEmailHandler":
                    admin_email_handler = handler
            if admin_email_handler in loggers.get(logger, {}).get("handlers", []):
                raise ValidationError({
                    "logger": _(
                        f"you cannot use logger: {logger} with handlers that contains "
                        f"'django.utils.log.AdminEmailHandler': {admin_email_handler}"
                    )
                })

    @staticmethod
    def remove_duplicates(address_list: list) -> list:
        return list(dict.fromkeys(address_list))

    def set_allowed_addresses(self):
        if self.allowed_addresses:
            self.allowed_addresses = self.remove_duplicates(self.allowed_addresses)
        else:
            self.allowed_addresses = []

    def set_forbidden_addresses(self):
        if self.forbidden_addresses:
            self.forbidden_addresses = self.remove_duplicates(self.forbidden_addresses)
        else:
            self.forbidden_addresses = []

    def set_allowed_domains(self):
        if self.allowed_domains:
            self.allowed_domains = self.remove_duplicates(self.allowed_domains)
        else:
            self.allowed_domains = []

    def set_forbidden_domains(self):
        if self.forbidden_domains:
            self.forbidden_domains = self.remove_duplicates(self.forbidden_domains)
        else:
            self.forbidden_domains = []

    def set_default_to_received_addresses(self):
        if self.default_to_received_addresses:
            self.default_to_received_addresses = self.remove_duplicates(self.default_to_received_addresses)
        else:
            self.default_to_received_addresses = []

    def set_default_cc_received_addresses(self):
        if self.default_cc_received_addresses:
            self.default_cc_received_addresses = self.remove_duplicates(self.default_cc_received_addresses)
        else:
            self.default_cc_received_addresses = []

    def set_default_bcc_received_addresses(self):
        if self.default_bcc_received_addresses:
            self.default_bcc_received_addresses = self.remove_duplicates(self.default_bcc_received_addresses)
        else:
            self.default_bcc_received_addresses = []

    def set_logger_level(self):
        if self.enable_loggers:
            if not self.logger_level:
                self.logger_level = logging._levelToName[logging.INFO]

    def set_loggers(self):
        if self.logger:
            self.logger = self.logger.lower()

    # def check_<profile>(self):
    #     """write here  check permissions """
    #     raise PermissionDenied

    def check_permissions(self):
        user = self.last_modifier
        if user.is_superuser:
            return
        # self.check_<profile>()
        # raise PermissionDenied("write here to override message")

    def set_status(self):
        """there can only be one instance enabled per environment"""
        if self.status == 1:
            email_backends = ThxEmailBackend.objects.filter(
                environment=self.environment, status=1
            ).exclude(name=self.name)
            email_backends.update(status=0)

    @staticmethod
    def _get_default_checks_error():
        msg = "EmailBackend cannot load correct settings. "
        msg2 = "Please set EMAIL_BACKEND_ENVIRONMENT correctly on your settings (" \
               "ex: production.py, staging.py, dev.py or base.py) with the your env name"
        return checks.Error(msg, hint=msg2), msg, msg2

    @classmethod
    def _check_required_settings(cls, **kwargs):
        error, msg, msg2 = cls._get_default_checks_error()
        try:
            env = settings.EMAIL_BACKEND_ENVIRONMENT
        except AttributeError as ae:
            setattr(error, 'msg', f"{ae} - {msg}")
            return [error]
        else:
            if not env or env == '< set in production.py | staging.py | dev.py >':
                return [error]
            if 'mail' not in settings.LOGGING['handlers']:
                # todo verificare perchè fatto così su
                return [checks.Error(
                    msg, hint="mail logger not found on LOGGING['handlers']. "
                    "Please set correct mail handler."
                )]
            if 'mail' not in settings.LOGGING['loggers']:
                return [checks.Error(
                    msg, hint="mail logger not found on LOGGING['loggers']. "
                    "Please set correct mail logger."
                )]
        return []

    @classmethod
    def _get_obj(cls, env_name: str):
        error, msg, msg2 = cls._get_default_checks_error()
        try:
            return cls.objects.get(environment=env_name.lower(), status=1)
        except MultipleObjectsReturned as mor:
            complete_msg = f"{mor} - EmailBacked is not properly set. - {msg2}"
        except cls.DoesNotExist as dne:
            complete_msg = f"{dne} - EmailBacked is not properly set. - {msg2}"
        logger.error(complete_msg)
        raise ValueError(complete_msg)

    @classmethod
    def check(cls, **kwargs):
        errors = super().check(**kwargs)
        errors.extend(cls._check_required_settings(**kwargs))
        return errors

    @classmethod
    def get_instance_form_environment(cls):
        errors = cls._check_required_settings()
        if errors:
            complete_msg = f"{errors[0].msg} - {errors[0].hint}"
            logger.error(complete_msg)
            raise ValueError(complete_msg)

        return cls._get_obj(settings.EMAIL_BACKEND_ENVIRONMENT)

    def _get_logger(self):
        import logging
        try:
            local_logger = logging.getLogger(self.logger)
            level = getattr(logging, self.logger_level.upper())
            local_logger.setLevel(level)
            local_logger.handlers[0].setLevel(level)
            return local_logger
        except Exception as e:
            logger.error(e)
            # TODO
            raise e

    def _log(self, message: str):
        if self.enable_loggers:
            logger = self._get_logger()
            lev = getattr(logger, self.logger_level.lower())
            return lev(message)
